import struct, math
from graphics import *

class Echo:
    # A simple class for xyz-points/vectors
    def __init__(self,x,y,z,intensity,picosecs):
        self.x = x
        self.y = y
        self.z = z
        self.intensity = intensity
        self.picosecs = picosecs

class Point3D:
    def __init__(self,x,y,z):
        self.x = x
        self.y = y
        self.z = z

class Pulse:
    def __init__(self, Gpstime, Echocount, Echoes,FileOffset, Poslidar, Stripnum, AGC, Wavetype, WF, Length):
        self.Gpstime = Gpstime
        self.Echocount = Echocount
        self.Echoes = Echoes
        self.FileOffset = FileOffset
        self.Poslidar = Poslidar
        self.Stripnum = Stripnum
        self.AGC = AGC
        self.Wavetype = Wavetype
        self.WF = WF
        self.Length = Length

class Point3D:
    # A simple class for xyz-points/vectors
    def __init__(self,x,y,z):
        self.x = x
        self.y = y
        self.z = z
    def norm(self):
        return math.sqrt(self.x**2+self.y**2+self.z**2)
    def normalize(self):  # * IN PLACE operation, normalize to len(1)
        length = self.norm()
        self.x = self.x / length
        self.y = self.y / length
        self.z = self.z / length


def Pituus(a):
        return math.sqrt((a.Echoes[0].x-a.Poslidar.x)**2+(a.Echoes[0].y-a.Poslidar.y)**2+(a.Echoes[0].z-a.Poslidar.z)**2)

def readHSVcolors(filename):
 # read 64 rgb colors, so called HSV-colors, RGB values in a CSV file
 infile = open(filename,"r")
 colors = [] #This list holds the values
 for line in infile:
    red = int(line.split(',')[0])
    green = int(line.split(',')[1])
    blue = int(line.split(',')[2])
    colors.append(color_rgb(red,green,blue))
    
 infile.close()
 return colors

def r_transform_matrix(Omega, Phi, Kappa):
  """ returns camera rotation matrix given omega, phi, kappa rotations"""
  # https://engineering.purdue.edu/~bethel/rot2.pdf
  COS = math.cos; SIN = math.sin; TAN = math.tan
  r = [[1,2,3,4],[1,2,3,4],[1,2,3,4],[1,2,3,4]]
  r[1][1] = COS(Phi) * COS(Kappa); r[1][2] = -COS(Phi) * SIN(Kappa); r[1][3]=SIN(Phi)
  r[2][1] = COS(Omega) * SIN(Kappa) + SIN(Omega) * SIN(Phi) * COS(Kappa)
  r[2][2] = COS(Omega) * COS(Kappa) - SIN(Omega) * SIN(Phi) * SIN(Kappa)
  r[2][3] =-SIN(Omega) * COS(Phi)
  r[3][1] = SIN(Omega) * SIN(Kappa) - COS(Omega) * SIN(Phi) * COS(Kappa)
  r[3][2] = SIN(Omega) * COS(Kappa) + COS(Omega) * SIN(Phi) * SIN(Kappa)
  r[3][3] = COS(Omega) * COS(Phi)
  return r

def r_transform_3D(X, Y, Z, c, r, cam_X, cam_Y, cam_Z):
    """ Calculates 3D --> 2D transformation, using collinear equations"""
    # INPUTS:
    # xyz   the point in 3d, in need of image position
    # c     camera constant (focal length), m
    # r     rotation matrix 3 x 3
    # cam_# camera's x, y and z.
    # OUTPUTS:
    # p_x, p_y the location of point xyz in the image (focal) plane, meters
    k = r[1][3]*(X - cam_X) + r[2][3]*(Y - cam_Y) + r[3][3]*(Z - cam_Z) 
    p_x = -c * (r[1][1] * (X - cam_X) + r[2][1] * (Y - cam_Y) + r[3][1] * (Z - cam_Z) ) / k
    p_y = -c * (r[1][2] * (X - cam_X) + r[2][2] * (Y - cam_Y) + r[3][2] * (Z - cam_Z) ) / k
    return p_x, p_y    

def transfrom_focal_plane(p_x,p_y, c):
    """Takes metric focal plane coordinates p_x, p_y and computes pixel position in image"""
    ImWidth =  800
    ImHeight = 800
    PixelWidth = 0.000005  # five micron pixel size
    col = ImWidth/2.0 + p_x / PixelWidth  # origin of p_x,p_y is in the middle pixel
    row = ImHeight/2.0 + p_y / PixelWidth
    return col, row

def readDEM(DemFileName):
    """Reads a raster DEM, returns it as a list"""
    f = open(DemFileName)
    lines = []
    while True:
      c = f.readline()
      c = c.rstrip("\n")
      if len(c) < 1 : break
      lines.append(c)
    Width = int(lines[0])
    Height = int(lines[1])
    Xll = float(lines[2])
    Yll = float(lines[3])
    CellSize = float(lines[4])
    NoData = float(lines[5])
    Filename = "C:\data\hyytiala_DEM_2010.bin" # lines[6]
    struct_fmt_DEM ='=' + str(Width) + "h" # short    
    struct_len_DEM = struct.calcsize(struct_fmt_DEM)
    struct_unpack_DEM = struct.Struct(struct_fmt_DEM).unpack_from
    print "Assisgning DEM list with zeros"
    s = [[0 for x in range(Height)] for y in range(Width)]
    print "Done. Now reading the DEM"
    with open(Filename, "rb") as f: # Avataan binaaritiedosto
        for row in range(Height):
           data = f.read(struct_len_DEM) # luetaan 2 tavun etumerkillinen kokonaisluku
           rowi = struct_unpack_DEM(data)
           for col in range(len(rowi)): 
             s[col][row] = rowi[col]*0.01   # puretaan se listaan, vain yksi alkio indeksilla nolla
           
    print "Done reading DEM (w x h)", Width, Height
    print s[1000][1000], type(s), type(s[1000][1000])
    return (s, Width, Height, Xll, Yll, CellSize, NoData)

DemFileName = "C:\data\Hyytiala_DEM_2010.hdr"
DEM, Width, Height, Xll, Yll, CellSize, NoData = readDEM(DemFileName)
print len(DEM),len(DEM[2079]), Width, Height, Xll, Yll, CellSize, NoData 
print DEM[1000][1000], type(DEM), type(DEM[1000][1000]), DEM[2079][4720]

struct_fmtA = '=i' # 4-tavun kokonaisluku, pulssien lukumaara bin-tiedostossa
struct_fmtB = '=dB12d4h5dfi5dh4B' # gps-time 0, pulsecount 1, 4 x xyz 2-4,5-7,8-10,11-13; 4 x intensity (14-17),
                                  # 4 x range (18-21), angle 22, roll 23, fileoffset 24
                                  # pitch 25, heading 26, poslidar xyz (27-29), stripnum (30), syncbit 31, AGC 32, res2 33, wavetype 34
struct_fmtWF = '=256B' # aaltomuoto 256 tavua (naytetta)
struct_lenA = struct.calcsize(struct_fmtA); struct_lenB = struct.calcsize(struct_fmtB)
struct_lenWF = struct.calcsize(struct_fmtWF) # rakenteiden pituus tavuina, lukemisen ohjaus
struct_unpackA = struct.Struct(struct_fmtA).unpack_from # maaritelma, kuinka tavut tietuessa muutetaan listan alkioiksi
struct_unpackB = struct.Struct(struct_fmtB).unpack_from
struct_unpackWF = struct.Struct(struct_fmtWF).unpack_from
Pulssit = []

MinX = 10E20; MinY=10E20;MaxX = 0; MaxY=0; MinZ = 10E20; MaxZ = -1000 # haetaan lidar-datan aariarvot, grafiikkaa varten
with open("f://markuskarppinen//052_098.bin", "rb") as f: # aletaan lukea bin-tiedostoa 
        data = f.read(struct_lenA) # luetaan 4 tavun etumerkillinen kokonaisluku
        s = struct_unpackA(data) # puretaan se listaan, vain yksi alkio indeksilla nolla
        NumofRecords = s[0]      # sijoitetaan
        print  str(NumofRecords) + " pulssia luettavaksi"               #
        for i in range(NumofRecords):
          data = f.read(struct_lenB)
          s = struct_unpackB(data)
          k = -1
          Kaiut = []
          for j in (3,2,1,0): # Assign xyz, intensity and time-difference for each echo, from 1 to 4 echoes, re-order 
             k += 1 
             if k + 1 <= s[1]: # Echocount in s[1]
               Kaiut.append(Echo(s[2+j*3], s[3+j*3], s[4+j*3], s[14+j], s[18+j]))
          Pulssit.append(Pulse(s[0], s[1],Kaiut, s[24], Point3D(s[27],s[28],s[29]), s[30],s[32],s[34], [], -99))
          #with open("f://markuskarppinen//052_098_256.bin", "rb") as g:
          #    g.seek(Pulssit[i].FileOffset,0)
          #    data = g.read(struct_lenWF)
          #    Pulssit[i].WF = struct_unpackWF(data)
          for k in range(Pulssit[i].Echocount):
            if Pulssit[i].Echoes[k].x < MinX: MinX = Pulssit[i].Echoes[k].x
            if Pulssit[i].Echoes[k].x > MaxX: MaxX = Pulssit[i].Echoes[k].x
            if Pulssit[i].Echoes[k].y < MinY: MinY = Pulssit[i].Echoes[k].y
            if Pulssit[i].Echoes[k].y > MaxY: MaxY = Pulssit[i].Echoes[k].y
            if Pulssit[i].Echoes[k].z < MinZ: MinZ = Pulssit[i].Echoes[k].z
            if Pulssit[i].Echoes[k].z > MaxZ: MaxZ = Pulssit[i].Echoes[k].z
         
        print "Luin " + str(i+1) + " pulssia"
        print MinX, MaxX, MinY, MaxY, MinZ, MaxZ

MaxZ = 180 # We have an outlier at 960 meters!

colors = readHSVcolors("f:/markuskarppinen/JetColorMAP.csv")
a = GraphWin("LiDAR data Visualizer", 800, 800, autoflush=False)
a.setBackground(color_rgb(255,255,255))
a.setCoords(-10.0,-10.0, 800.0, 800.0)
msg = Text(Point(50,50),"Click to begin")
msg.draw(a)
#w=Image(Point(400,300),"f:/markuskarppinen/kuvapala.ppm")
#w.draw(a)
a.getMouse()
c = 0.004 # polttovali
Omega = 0.1;  Phi = -0.00004;  Kappa = -0.00 # kameran kierrot
r = r_transform_matrix(Omega, Phi, Kappa) # kiertomatriisi
cam_X = MinX+50;   cam_Y = MinY+25;   cam_Z = 300 # kameran sijainti
for i in range(NumofRecords):
    for j in range(Pulssit[i].Echocount):
        P = Point3D(0,0,0)
        P.x = Pulssit[i].Echoes[j].x; P.y = Pulssit[i].Echoes[j].y; P.z = Pulssit[i].Echoes[j].z
        # image coordinates
        PlotDEM = True
        col = (P.x - MinX)*5
        row = (P.y - MinY)*5
        # We need UTM (ETRS-TM35FIN) coords with the DEM
        UTM_E = -2471441.562 + 0.9987798071 * P.x + 0.04612734592 * P.y
        UTM_N = 124518.3273 - 0.04613846192 * P.x + 0.9987750048 * P.y
        #dem-coordinates
        col_dem = int((UTM_E-Xll)/CellSize)
        row_dem = Height-int((UTM_N-Yll)/CellSize)
   #     print P.x, P.y, UTM_E, UTM_N, Xll, Yll, col_dem, row_dem
        if PlotDEM == True:
            MinZ=0
            MaxZ=20
            zgnd = DEM[col_dem][row_dem] - 18.67 + 0.32 # N60-elliosoid height difference + N60-N2000 difference (zgnd==N2000 elevation)
            h = P.z-zgnd
            colorj = int((1-(MaxZ-h)/(MaxZ-MinZ))*63)
        if PlotDEM == False:
            colorj = int((1-(MaxZ-P.z)/(MaxZ-MinZ))*63)
        if colorj > 63: colorj = 63 # Don't let the index overflow
        color = colors[colorj] # Finally select the color
#        a.plot(int(col),int(row),color)
#        if (i % 20000) == 0: a.update()

        # camera parameters, focal length, rotations, position
 
        p_x, p_y = r_transform_3D(P.x , P.y, P.z, c, r, cam_X, cam_Y, cam_Z)
        col, row = transfrom_focal_plane(p_x,p_y, c)
        a.plot(int(col),int(row),color)
        if (i % 20000) == 0: a.update()
   

        
        
msg = Text(Point(100,100),"Click to quit")
msg.draw(a)
a.getMouse()
  
a.close()

